# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for NICs.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS
from ._cmd_partition import find_partition
from ._cmd_cpc import find_cpc


# Defaults for NIC creation
SSC_IP_ADDRESS_TYPES = ['ipv4', 'ipv6', 'linklocal', 'dhcp']
VLAN_TYPES = ['enforced', 'none']


def find_nic(cmd_ctx, client, cpc_or_name, partition_name, nic_name):
    """
    Find a NIC by name and return its resource object.
    """
    partition = find_partition(cmd_ctx, client, cpc_or_name, partition_name)
    try:
        nic = partition.nics.find(name=nic_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return nic


@cli.group('nic', options_metavar=COMMAND_OPTIONS_METAVAR)
def nic_group():
    """
    Command group for managing NICs (DPM mode only).

    The commands in this group work only on CPCs that are in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@nic_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--type', is_flag=True, required=False, hidden=True)
@add_options(LIST_OPTIONS)
@click.pass_obj
def nic_list(cmd_ctx, cpc, partition, **options):
    """
    List the NICs in a partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_list(cmd_ctx, cpc, partition, options))


@nic_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.pass_obj
def nic_show(cmd_ctx, cpc, partition, nic):
    """
    Show the details of a NIC.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Partition.
      - 'virtual-switch-name' - Name of the Virtual Switch for the backing
        OSA/Hipersocket Adapter
      - 'network-adapter-name' - Name of the backing Adapter
      - 'network-adapter-port-name' - Name of the backing Adapter Port
      - 'network-adapter-port-index' - Index of the backing Adapter Port

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_show(cmd_ctx, cpc, partition, nic))


@nic_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--name', type=str, required=True,
              help='The name of the new NIC.')
@click.option('--description', type=str, required=False,
              help='The description of the new NIC. '
              'Default: empty')
@click.option('--adapter', type=str, required=False,
              help='The name of the network adapter with the port backing the '
              'new NIC. Required.')
@click.option('--port', type=str, required=False,
              help='The name or index of the network port backing the new NIC. '
              'Required.')
@click.option('--virtual-switch', type=str, required=False,
              help='Deprecated: The name of the virtual switch of the network '
              'port backing the new NIC. Use --adapter and --port instead.')
@click.option('--device-number', type=str, required=False,
              help='The device number to be used for the new NIC. '
              'Default: auto-generated')
@click.option('--ssc-management-nic', type=bool, required=False,
              help='Indicates that this NIC should be used as a management '
              'NIC for Secure Service Container to access the web interface. '
              'Only applicable to NICs of SSC partitions. '
              'Default: False')
@click.option('--ssc-ip-address-type', type=click.Choice(SSC_IP_ADDRESS_TYPES),
              required=False,
              help='Secure Service Container IP address type. '
              'Only applicable to and required for NICs of '
              'SSC partitions.')
@click.option('--ssc-ip-address', type=str, required=False,
              help='IP Address of the SSC management web interface. '
              'Only applicable to and required for NICs of SSC '
              'partitions when ssc-ip-address-type is ipv4 or ipv6.')
@click.option('--ssc-mask-prefix', type=str, required=False,
              help='Network Mask of the SSC management NIC. '
              'Only applicable to and required for NICs of SSC '
              'partitions when ssc-ip-address-type is ipv4 or ipv6.')
@click.option('--vlan-id', metavar='[INT|none]', type=str, required=False,
              help='VLAN ID of the NIC, or "none" for setting no VLAN ID. '
              'On z14 or later CPCs, specifying a VLAN ID requires using the '
              '--vlan-type option. '
              'Only applicable to management NICs of SSC partitions, and to '
              'OSA and Hipersocket NICs of non-SSC partitions. '
              'Default: No VLAN ID')
@click.option('--vlan-type', required=False, type=click.Choice(VLAN_TYPES),
              help='VLAN type of the NIC, or "none" for setting no VLAN type. '
              'Only supported on z14 or later CPCs, and required if specifying '
              'a VLAN ID with the --vlan-id option. '
              'Default: No VLAN type')
@click.pass_obj
def nic_create(cmd_ctx, cpc, partition, **options):
    """
    Create a NIC in a partition.

    The NIC is backed by a port (jack) on an adapter. For all types of network
    adapters, the backing adapter and port can be specified with the --adapter
    and --port options. The --virtual-switch option is deprecated but still
    supported for compatibility; it can be used only for OSA and HiperSocket
    adapters.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_create(cmd_ctx, cpc, partition,
                                               options))


@nic_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.option('--name', type=str, required=False,
              help='The new name of the NIC.')
@click.option('--description', type=str, required=False,
              help='The new description of the NIC.')
@click.option('--adapter', type=str, required=False,
              help='The name of the network adapter with the port backing the '
              'NIC. Required.')
@click.option('--port', type=str, required=False,
              help='The name or index of the network port backing the NIC. '
              'Required.')
@click.option('--virtual-switch', type=str, required=False,
              help='Deprecated: The name of the virtual switch of the network '
              'port backing the NIC. Use --adapter and --port instead.')
@click.option('--device-number', type=str, required=False,
              help='The new device number to be used for the NIC.')
@click.option('--ssc-management-nic', type=bool, required=False,
              help='Indicates that this NIC should be used as a management '
              'NIC for Secure Service Container to access the web interface. '
              'Only applicable to NICs of SSC partitions. ')
@click.option('--ssc-ip-address-type', type=click.Choice(SSC_IP_ADDRESS_TYPES),
              required=False,
              help='Secure Service Container IP address type. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--ssc-ip-address', type=str, required=False,
              help='IP Address of the SSC management web interface. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--ssc-mask-prefix', type=str, required=False,
              help='Network Mask of the SSC management NIC. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--vlan-id', metavar='[INT|none]', type=str, required=False,
              help='VLAN ID of the NIC, or "none" for setting no VLAN ID. '
              'On z14 or later CPCs, specifying a VLAN ID requires using the '
              '--vlan-type option. '
              'Only applicable to management NICs of SSC partitions, and to '
              'OSA and Hipersocket NICs of non-SSC partitions.')
@click.option('--vlan-type', required=False, type=click.Choice(VLAN_TYPES),
              help='VLAN type of the NIC, or "none" for setting no VLAN type. '
              'Only supported on z14 or later CPCs, and required if specifying '
              'a VLAN ID with the --vlan-id option.')
@click.pass_obj
def nic_update(cmd_ctx, cpc, partition, nic, **options):
    """
    Update the properties of a NIC.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_update(cmd_ctx, cpc, partition, nic,
                                               options))


@nic_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the NIC.',
              prompt='Are you sure you want to delete this NIC ?')
@click.pass_obj
def nic_delete(cmd_ctx, cpc, partition, nic):
    """
    Delete a NIC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_delete(cmd_ctx, cpc, partition, nic))


def backing_uri(cmd_ctx, cpc, org_options, required=False):
    """
    Determine the backing adapter port or vswitch to be used from the
    --adapter, --port, and --virtual-switch options, and return a dict with the
    correct property for the URI of the backing port or vswitch to be used for
    the "Create NIC" operation.

    Returns:
      dict with the backing URI property set, if backing object specified
      None, if no backing object specified and not required

    Raises:
      click exception for various error situations.
    """

    if org_options['virtual-switch']:
        # This option is deprecated, but it is still supported for
        # backwards compatibility.

        if org_options['adapter'] or org_options['port']:
            raise click_exception(
                "The (deprecated) --virtual-switch option must not be "
                "specified together with any of the --adapter or --port "
                "options.",
                cmd_ctx.error_format)

        vswitch_name = org_options['virtual-switch']
        try:
            vswitch = cpc.virtual_switches.find(name=vswitch_name)
        except zhmcclient.NotFound:
            raise click_exception(
                "Could not find virtual switch '{s}' in CPC '{c}'.".
                format(s=vswitch_name, c=cpc.name),
                cmd_ctx.error_format)
        return {'virtual-switch-uri': vswitch.uri}

    if bool(org_options['adapter']) != bool(org_options['port']):
        raise click_exception(
            "The --adapter and --port options must be specified both or none.",
            cmd_ctx.error_format)

    if org_options['adapter']:
        adapter_name = org_options['adapter']
    else:
        if required:
            raise click_exception(
                "Required --adapter option is not specified",
                cmd_ctx.error_format)
        return None

    if org_options['port']:
        port_name = org_options['port']
    else:
        if required:
            raise click_exception(
                "Required --port option is not specified",
                cmd_ctx.error_format)
        return None

    try:
        adapter = cpc.adapters.find(name=adapter_name)
    except zhmcclient.NotFound:
        raise click_exception(
            "Could not find adapter '{a}' in CPC '{c}'.".
            format(a=adapter_name, c=cpc.name),
            cmd_ctx.error_format)

    try:
        port = adapter.ports.find(name=port_name)
    except zhmcclient.NotFound:
        # Try interpreting the --port value as a port index
        try:
            port_index = int(port_name)
        except ValueError:
            raise click_exception(
                "Could not find port with name '{p}' on "
                "adapter '{a}' in CPC '{c}'.".
                format(p=port_name, a=adapter_name, c=cpc.name),
                cmd_ctx.error_format)
        try:
            port = adapter.ports.find(index=port_index)
        except zhmcclient.NotFound:
            raise click_exception(
                "Could not find port with name or index '{p}' on "
                "adapter '{a}' in CPC '{c}'.".
                format(p=port_name, a=adapter_name, c=cpc.name),
                cmd_ctx.error_format)

    adapter_type = adapter.get_property('type')
    if adapter_type in ('roce', 'cna'):
        return {'network-adapter-port-uri': port.uri}

    if adapter_type in ('osd', 'hipersockets'):
        port_index = port.get_property('index')
        filter_args = {
            'backing-adapter-uri': adapter.uri,
            'port': port_index,
        }
        try:
            vswitch = cpc.virtual_switches.find(**filter_args)
        except zhmcclient.NotFound:
            raise click_exception(
                "Could not find virtual switch with backing adapter '{a}' "
                "and port index '{p}' in CPC '{c}'.".
                format(a=adapter.name, p=port_index, c=cpc.name),
                cmd_ctx.error_format)
        return {'virtual-switch-uri': vswitch.uri}

    raise click_exception(
        "Adapter '{a}' on CPC '{c}' has unsupported type {t} for "
        "being a backing adapter of a NIC.".
        format(a=adapter_name, c=cpc.name, t=adapter_type),
        cmd_ctx.error_format)


def cmd_nic_list(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    try:
        nics = partition.nics.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['type']:
        click.echo("The --type option is deprecated and type information "
                   "is now always shown.")

    show_list = [
        'name',
        'cpc',
        'partition',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    partition_additions = {}
    for nic in nics:
        cpc_additions[nic.uri] = cpc_name
        partition_additions[nic.uri] = partition_name
    additions = {
        'cpc': cpc_additions,
        'partition': partition_additions,
    }

    try:
        print_resources(cmd_ctx, nics, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_nic_show(cmd_ctx, cpc_name, partition_name, nic_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    try:
        nic.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(nic.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = partition_name

    # Add artificial properties in case of a vswitch-based NIC (OSA, HS)
    try:
        vswitch_uri = nic.get_property('virtual-switch-uri')
    except KeyError:
        pass
    else:
        vswitch_props = client.session.get(vswitch_uri)
        properties['virtual-switch-name'] = vswitch_props['name']

        cpc = find_cpc(cmd_ctx, client, cpc_name)
        adapter_uri = vswitch_props['backing-adapter-uri']
        adapter = cpc.adapters.resource_object(adapter_uri)
        properties['network-adapter-name'] = adapter.name

        port_index = vswitch_props['port']
        properties['network-adapter-port-index'] = port_index

        port = adapter.ports.find(index=port_index)
        properties['network-adapter-port-name'] = port.name

    # Add artificial properties in case of an adapter-based NIC (RoCE, CNA)
    try:
        port_uri = nic.get_property('network-adapter-port-uri')
    except KeyError:
        pass
    else:
        port_props = client.session.get(port_uri)
        properties['network-adapter-port-name'] = port_props['name']
        properties['network-adapter-port-index'] = port_props['index']

        adapter_props = client.session.get(port_props['parent'])
        properties['network-adapter-name'] = adapter_props['name']

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_nic_create(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
        'port': None,
        'virtual-switch': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    set_vlan_id_type(cmd_ctx, properties, org_options)

    properties.update(backing_uri(
        cmd_ctx, partition.manager.cpc, org_options, required=True))

    try:
        new_nic = partition.nics.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New NIC '{n}' has been created.".
               format(n=new_nic.properties['name']))


def cmd_nic_update(cmd_ctx, cpc_name, partition_name, nic_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
        'port': None,
        'virtual-switch': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    set_vlan_id_type(cmd_ctx, properties, org_options)

    uri_prop = backing_uri(
        cmd_ctx, nic.manager.partition.manager.cpc, org_options)
    if uri_prop:
        properties.update(uri_prop)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating NIC '{n}'.".
                   format(n=nic_name))
        return

    try:
        nic.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != nic_name:
        click.echo("NIC '{n}' has been renamed to '{nn}' and was updated.".
                   format(n=nic_name, nn=properties['name']))
    else:
        click.echo("NIC '{n}' has been updated.".format(n=nic_name))


def cmd_nic_delete(cmd_ctx, cpc_name, partition_name, nic_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    try:
        nic.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("NIC '{n}' has been deleted.".format(n=nic_name))


def set_vlan_id_type(cmd_ctx, properties, org_options):
    """
    Set the 'vlan-id' and 'vlan-type' properties from the options.
    """
    vlan_id = org_options['vlan-id']
    if vlan_id is None:
        pass  # No change on update, use HMC default on create
    elif vlan_id == 'none':
        # Reset to no VLAN ID (important for update)
        properties['vlan-id'] = None
    else:
        try:
            properties['vlan-id'] = int(vlan_id)
        except ValueError:
            raise click_exception(
                "Invalid value for '--vlan-id': {} is not a valid integer".
                format(vlan_id), cmd_ctx.error_format)

    vlan_type = org_options['vlan-type']
    if vlan_type is None:
        pass  # No change on update, use HMC default on create
    elif vlan_type == 'none':
        # Reset to no VLAN type (important for update)
        properties['vlan-type'] = None
    else:
        properties['vlan-type'] = vlan_type



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC user management.
"""

from __future__ import absolute_import
from __future__ import print_function

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, ObjectByUriCache, \
    API_VERSION_HMC_2_14_0, API_VERSION_HMC_2_15_0


def find_user(cmd_ctx, console, user_name):
    """
    Find a user by name and return its resource object.
    """
    try:
        user = console.users.find(name=user_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return user


@cli.group('user', options_metavar=COMMAND_OPTIONS_METAVAR)
def user_group():
    """
    Command group for managing HMC users.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@user_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@add_options(LIST_OPTIONS)
@click.option('--permissions', is_flag=True, required=False,
              help='Show additional properties for user permissions and roles.')
@click.option('--status', is_flag=True, required=False,
              help='Show additional properties for user status.')
@click.pass_obj
def user_list(cmd_ctx, **options):
    """
    List the HMC users.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_list(cmd_ctx, options))


@user_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.pass_obj
def user_show(cmd_ctx, user):
    """
    Show the details of an HMC user.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Console.
      - 'user-role-names' - Names of the User Roles referenced by
        'user-role-uris' (index-correlated).
      - 'password-rule-name' - Name of the Password Rule referenced by
        'password-rule-uri'.
      - 'user-pattern-name' - Name of the User Pattern referenced by
        'user-pattern-uri', if present (for pattern-based users)
      - 'user-template-name' - Name of the User Template referenced by
        'user-template-uri', if present (for pattern-based users)
      - 'default-group-name' - Name of the Group referenced by
        'default-group-uri'.
      - 'ldap-server-definition-name' - Name of the LDAP Server Definition
        referenced by 'ldap-server-definition-uri'.
      - 'primary-mfa-server-definition-name' - Name of the MFA Server Definition
        referenced by 'primary-mfa-server-definition-uri', if present (since
        HMC 2.15.0).
      - 'backup-mfa-server-definition-name' - Name of the MFA Server Definition
        referenced by 'backup-mfa-server-definition-uri', if present (since
        HMC 2.15.0).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_show(cmd_ctx, user))


@user_group.command('password', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.pass_obj
def user_password(cmd_ctx, user):
    """
    Set a new password for an HMC user.

    The new password is prompted for.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_password(cmd_ctx, user))


@user_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, required=True,
              help='The name of the new user.')
@click.option('--like', type=str, required=False,
              help='The name of a like user. The properties of the like user '
              'will be used as defaults for the new user, except for the '
              'following properties: name, password, description, '
              'email-address, mfa-userid, mfa-userid-override. These defaults '
              'can be overridden using options.')
@click.option('--type', type=click.Choice(['standard', 'template']),
              required=False, default='standard',
              help='The type of the new user. '
              'Default: standard')
@click.option('--description', type=str, required=False,
              help='The description of the new user.')
@click.option('--email-address', type=str, required=False,
              help='The email address of the new user, or the empty string to '
              'set no email address. '
              'Requires HMC 2.14.0 or later. '
              'Default: No email address')
@click.option('--disabled', type=bool, required=False,
              help='The disabled state of the new user. '
              'Default: False')
@click.option('--authentication-type', type=click.Choice(['local', 'ldap']),
              required=False, default='local',
              help='The authentication type of the new user. '
              'Default: local')
@click.option('--password-rule', type=str, required=False,
              help='The name of the password rule of the new user. '
              'Only valid and required with authentication type "local".')
@click.option('--password', type=str, required=False,
              help='The logon password for the new user. '
              'Only valid and required with authentication type "local".')
@click.option('--force-password-change', type=bool, required=False,
              help='The force-password-change state of the new user. '
              'Only valid with authentication type "local". '
              'Default: False')
@click.option('--ldap-server-definition', type=str, required=False,
              help='The name of the LDAP server definition of the new user, or '
              'the empty string to set no LDAP server definition. '
              'Only valid and required with authentication type "ldap".')
@click.option('--userid-on-ldap-server', type=str, required=False,
              help='The userid on the LDAP server. '
              'Only valid with authentication type "ldap".')
@click.option('--session-timeout', type=int, required=False,
              help='The session timeout in minutes for the new user. '
              'This is the amount of time for which a user\'s UI session can '
              'run before being prompted for identity verification. '
              '0 indicates no timeout. '
              'Default: 0')
@click.option('--verify-timeout', type=int, required=False,
              help='The verification timeout in minutes for the new user. '
              'This is the amount of time allowed for the user to re-enter '
              'their password when being prompted due to a session timeout. '
              '0 indicates no timeout. '
              'Default: 15')
@click.option('--idle-timeout', type=int, required=False,
              help='The idle timeout in minutes for the new user. '
              'This is the amount of time the user\'s UI session can be idle '
              'before it is disconnected. '
              '0 indicates no timeout. '
              'Default: 0')
@click.option('--min-pw-change-time', type=int, required=False,
              help='The minimum password change time in minutes for the new '
              'user. This is the minimum amount of time that must pass between '
              'changes to the user\'s password. '
              '0 indicates no minimum time. '
              'Only valid with authentication type "local". '
              'Default: 0')
@click.option('--max-failed-logins', type=int, required=False,
              help='The maximum number of consecutive failed login attempts '
              'for the new user. When exceeding this maximum, the user '
              'will be temporarily disabled for the amount of time specified '
              'in the "disable-delay" property. '
              '0 indicates that the user is never disabled due to failed '
              'login attempts. '
              'Default: 3')
@click.option('--disable-delay', type=int, required=False,
              help='The disable-delay time in minutes for the new user. '
              'This is the amount of time the user will be temporarily '
              'disabled after exceeding the maximum number of failed login '
              'attempts. '
              '0 indicates that the user is not disabled for any period of '
              'time after reaching the maximum number of invalid login '
              'attempts. '
              'Default: 1')
@click.option('--inactivity-timeout', type=int, required=False,
              help='The inactivity timeout in days for the new user. '
              'This is the maximum number of consecutive days of with no login '
              'before the user is disabled. '
              '0 indicates no inactivity timeout. '
              'Default: 0')
@click.option('--disruptive-pw-required', type=bool, required=False,
              help='The indicator whether the new user\'s password is required '
              'to perform disruptive actions through the UI. '
              'Default: True')
@click.option('--disruptive-text-required', type=bool, required=False,
              help='The indicator whether text input is required to '
              'perform disruptive actions through the UI. '
              'Default: False')
@click.option('--allow-remote-access', type=bool, required=False,
              help='The indicator whether the new user is allowed to access '
              'the HMC through its remote web server interface. '
              'Default: False')
@click.option('--allow-management-interfaces', type=bool, required=False,
              help='The indicator whether the new user is allowed access to '
              'management interfaces. This includes access to the '
              'Web Services APIs. '
              'Default: False')
@click.option('--max-web-services-api-sessions', type=int, required=False,
              help='The maximum number of simultaneous Web Services API '
              'sessions the new user is permitted to have. '
              'Default: 100')
@click.option('--web-services-api-session-idle-timeout', type=int,
              required=False,
              help='The idle timeout in minutes for Web Services API sessions '
              'created by the new user. This is the amount of time a Web '
              'Services API session can be idle before it is terminated. '
              'Default: 360')
@click.option('--mfa-type', type=click.Choice(['hmc-totp', 'mfa-server', '']),
              required=False,
              help='The MFA type of the new user, or the empty string for '
              'no MFA. '
              'Requires HMC 2.15.0 or later. '
              'Default: No MFA')
@click.option('--primary-mfa-server-definition', type=str, required=False,
              help='The name of the MFA Server Definition for the primary MFA '
              'server used to authenticate the new user, or the empty string '
              'to set no such server. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later. '
              'Default: No such server')
@click.option('--backup-mfa-server-definition', type=str, required=False,
              help='The name of the MFA Server Definition for the backup MFA '
              'server used to authenticate the new user, or the empty string '
              'to set no such server. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later. '
              'Default: No such server')
@click.option('--mfa-policy', type=str, required=False,
              help='The name of the MFA policy for the new user, or the empty '
              'string to set no MFA policy. This is for example a RACF policy. '
              'The MFA policy applies to the user when an MFA server '
              'authenticates the user. It must identify a policy whose only '
              'MFA factor is the RSA SecurID factor. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later. '
              'Default: No MFA policy')
@click.option('--mfa-userid', type=str, required=False,
              help='The name of the MFA user ID for the new user, or the empty '
              'string to set no MFA user ID. This is a user ID, such as a RACF '
              'user ID, that identifies this user to the MFA server that '
              'authenticates the user. '
              'Only valid for MFA type "mfa-server" and user type not '
              '"template". '
              'Requires HMC 2.15.0 or later. '
              'Default: No MFA user ID')
@click.option('--mfa-userid-override', type=str, required=False,
              help='The name of the LDAP attribute that contains the MFA '
              'user ID that overrides the mfa-userid during authentication, or '
              'the empty string to set no userid override via LDAP attribute. '
              'Only valid for MFA type "mfa-server" and user type "template". '
              'Requires HMC 2.15.0 or later. '
              'Default: No userid override')
@click.pass_obj
def user_create(cmd_ctx, **options):
    """
    Create an HMC user.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_create(cmd_ctx, options))


@user_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.option('--description', type=str, required=False,
              help='The new description of the user.')
@click.option('--email-address', type=str, required=False,
              help='The new email address of the user or the empty string to '
              'set no email address. '
              'Requires HMC 2.14.0 or later.')
@click.option('--disabled', type=bool, required=False,
              help='The new disabled state of the user.')
@click.option('--password-rule', type=str, required=False,
              help='The name of the new password rule of the user. '
              'Only valid with authentication type "local".')
@click.option('--password', type=str, required=False,
              help='The new logon password for the user.')
@click.option('--force-password-change', type=bool, required=False,
              help='The new force-password-change state of the user. '
              'Only valid with authentication type "local".')
@click.option('--default-group', type=str, required=False,
              help='The name of the new default group of the user, or the '
              'empty string to set no default group. '
              'Managed objects created by the user automatically become '
              'members of its default group, if set.')
@click.option('--authentication-type', type=click.Choice(['local', 'ldap']),
              required=False,
              help='The new authentication type of the user.')
@click.option('--ldap-server-definition', type=str, required=False,
              help='The name of the new LDAP server definition of the user, or '
              'the empty string to set no LDAP server definition. '
              'Only valid with authentication type "ldap".')
@click.option('--userid-on-ldap-server', type=str, required=False,
              help='The new userid on the LDAP server. '
              'Only valid with authentication type "ldap".')
@click.option('--session-timeout', type=int, required=False,
              help='The new session timeout in minutes for the user. '
              'This is the amount of time for which a user\'s UI session can '
              'run before being prompted for identity verification. '
              '0 indicates no timeout.')
@click.option('--verify-timeout', type=int, required=False,
              help='The new verification timeout in minutes for the user. '
              'This is the amount of time allowed for the user to re-enter '
              'their password when being prompted due to a session timeout. '
              '0 indicates no timeout.')
@click.option('--idle-timeout', type=int, required=False,
              help='The new idle timeout in minutes for the user. '
              'This is the amount of time the user\'s UI session can be idle '
              'before it is disconnected. '
              '0 indicates no timeout.')
@click.option('--min-pw-change-time', type=int, required=False,
              help='The new minimum password change time in minutes for the '
              'user. This is the minimum amount of time that must pass between '
              'changes to the user\'s password. '
              '0 indicates no minimum time. '
              'Only valid with authentication type "local".')
@click.option('--max-failed-logins', type=int, required=False,
              help='The new maximum number of consecutive failed login '
              'attempts for the user. When exceeding this maximum, the user '
              'will be temporarily disabled for the amount of time specified '
              'in the "disable-delay" property. '
              '0 indicates that the user is never disabled due to failed '
              'login attempts.')
@click.option('--disable-delay', type=int, required=False,
              help='The new disable-delay time in minutes for the user. '
              'This is the amount of time the user will be temporarily '
              'disabled after exceeding the maximum number of failed login '
              'attempts. '
              '0 indicates that the user is not disabled for any period of '
              'time after reaching the maximum number of invalid login '
              'attempts.')
@click.option('--inactivity-timeout', type=int, required=False,
              help='The new inactivity timeout in days for the user. '
              'This is the maximum number of consecutive days of with no login '
              'before the user is disabled. '
              '0 indicates no inactivity timeout.')
@click.option('--disruptive-pw-required', type=bool, required=False,
              help='The new indicator whether the user\'s password is required '
              'to perform disruptive actions through the UI.')
@click.option('--disruptive-text-required', type=bool, required=False,
              help='The new indicator whether text input is required to '
              'perform disruptive actions through the UI.')
@click.option('--allow-remote-access', type=bool, required=False,
              help='The new indicator whether the user is allowed to access '
              'the HMC through its remote web server interface.')
@click.option('--allow-management-interfaces', type=bool, required=False,
              help='The new indicator whether the user is allowed access to '
              'management interfaces. This includes access to the '
              'Web Services APIs.')
@click.option('--max-web-services-api-sessions', type=int, required=False,
              help='The new maximum number of simultaneous Web Services API '
              'sessions the user is permitted to have.')
@click.option('--web-services-api-session-idle-timeout', type=int,
              required=False,
              help='The new idle timeout in minutes for Web Services API '
              'sessions created by the user. This is the amount of time a Web '
              'Services API session can be idle before it is terminated.')
@click.option('--mfa-type', type=click.Choice(['hmc-totp', 'mfa-server', '']),
              required=False,
              help='The new MFA type of the user, or the empty string for '
              'no MFA. '
              'Requires HMC 2.15.0 or later.')
@click.option('--force-shared-secret-key-change', type=bool, required=False,
              help='The new indicator whether the user is required to '
              'establish a new shared secret key during the next logon. The '
              'shared secret key is used to calculate the user\'s current TOTP '
              'multi-factor authentication code. '
              'Requires HMC 2.14.0 or later.')
@click.option('--primary-mfa-server-definition', type=str, required=False,
              help='The name of the new MFA Server Definition for the primary '
              'MFA server used to authenticate the user, or the empty string '
              'to set no such server. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later.')
@click.option('--backup-mfa-server-definition', type=str, required=False,
              help='The name of the new MFA Server Definition for the backup '
              'MFA server used to authenticate the user, or the empty string '
              'to set no such server. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later.')
@click.option('--mfa-policy', type=str, required=False,
              help='The name of the new MFA policy for the user, or the empty '
              'string to set no MFA policy. This is for example a RACF policy. '
              'The MFA policy applies to the user when an MFA server '
              'authenticates the user. It must identify a policy whose only '
              'MFA factor is the RSA SecurID factor. '
              'Only valid for MFA type "mfa-server". '
              'Requires HMC 2.15.0 or later.')
@click.option('--mfa-userid', type=str, required=False,
              help='The name of the new MFA user ID for the user, or the empty '
              'string to set no MFA user ID. This is a user ID, such as a RACF '
              'user ID, that identifies this user to the MFA server that '
              'authenticates the user. '
              'Only valid for MFA type "mfa-server" and user type not '
              '"template". '
              'Requires HMC 2.15.0 or later.')
@click.option('--mfa-userid-override', type=str, required=False,
              help='The name of the new LDAP attribute that contains the MFA '
              'user ID that overrides the mfa-userid during authentication, or '
              'the empty string to set no userid override via LDAP attribute. '
              'Only valid for MFA type "mfa-server" and user type "template". '
              'Requires HMC 2.15.0 or later.')
@click.pass_obj
def user_update(cmd_ctx, user, **options):
    """
    Update the properties of an HMC user.

    Note that HMC users cannot be renamed.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_update(cmd_ctx, user, options))


@user_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the user.',
              prompt='Are you sure you want to delete this user ?')
@click.pass_obj
def user_delete(cmd_ctx, user):
    """
    Delete an HMC user.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_delete(cmd_ctx, user))


@user_group.command('add-role', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@click.pass_obj
def user_add_role(cmd_ctx, user, user_role):
    """
    Add a user role to a user.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_add_role(cmd_ctx, user, user_role))


@user_group.command('remove-role', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER', type=str, metavar='USER')
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@click.pass_obj
def user_remove_role(cmd_ctx, user, user_role):
    """
    Remove a user role from a user.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_remove_role(cmd_ctx, user, user_role))


def cmd_user_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'email-address',
            'description',
            'type',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])
    if options['permissions']:
        show_list.extend([
            'allow-management-interfaces',
            'allow-remote-access',
            'roles',  # addition
        ])
    if options['status']:
        show_list.extend([
            'disabled',
            'is-locked',
            'max-failed-logins',
            'password-expires',
            'password-rule',  # addition
            'force-password-change',
            'multi-factor-authentication-required',
            'force-shared-secret-key-change',
        ])

    additions = {}
    additions['roles'] = {}
    additions['password-rule'] = {}

    try:
        users = console.users.list(full_properties=False)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    obj_cache = None

    if options['permissions']:
        if obj_cache is None:
            obj_cache = ObjectByUriCache(cmd_ctx, client)
        for user in users:
            role_names = [obj_cache.user_role_by_uri(role_uri).name
                          for role_uri in user.get_property('user-roles')]
            additions['roles'][user.uri] = role_names

    if options['status']:
        if obj_cache is None:
            obj_cache = ObjectByUriCache(cmd_ctx, client)
        for user in users:
            rule_uri = user.get_property('password-rule-uri')
            if rule_uri:
                rule_name = obj_cache.password_rule_by_uri(rule_uri).name
                additions['password-rule'][user.uri] = rule_name
            else:
                additions['password-rule'][user.uri] = None

    try:
        print_resources(cmd_ctx, users, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_user_show(cmd_ctx, user_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)

    try:
        user.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(user.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = console.name

    obj_cache = ObjectByUriCache(cmd_ctx, client)

    # Add artificial property 'user-role-names'
    role_names = [obj_cache.user_role_by_uri(role_uri).name
                  for role_uri in user.properties['user-roles']]
    properties['user-role-names'] = role_names

    # Add artificial property 'password-rule-name'
    rule_uri = user.properties['password-rule-uri']
    if rule_uri:
        # Authentication type is local
        rule_name = obj_cache.password_rule_by_uri(rule_uri).name
        properties['password-rule-name'] = rule_name
    else:
        # Authentication type is LDAP
        properties['password-rule-name'] = None

    # Add artificial property 'user-pattern-name'
    try:
        upat_uri = user.get_property('user-pattern-uri')
    except KeyError:
        pass
    else:
        # User is pattern-based
        upat_props = client.session.get(upat_uri)
        properties['user-pattern-name'] = upat_props['name']

    # Add artificial property 'user-template-name'
    try:
        utemp_uri = user.get_property('user-template-uri')
    except KeyError:
        pass
    else:
        # User is pattern-based
        utemp_props = client.session.get(utemp_uri)
        properties['user-template-name'] = utemp_props['name']

    # Add artificial property 'default-group-name'
    defgrp_uri = user.get_property('default-group-uri')
    if defgrp_uri:
        # User has a default group
        defgrp_props = client.session.get(defgrp_uri)
        properties['default-group-name'] = defgrp_props['name']
    else:
        # User has no default group
        properties['default-group-name'] = None

    # Add artificial property 'ldap-server-definition-name'
    lsd_uri = user.get_property('ldap-server-definition-uri')
    if lsd_uri:
        # User authentication type is LDAP
        lsd_props = client.session.get(lsd_uri)
        properties['ldap-server-definition-name'] = lsd_props['name']
    else:
        # User authentication type is local
        properties['ldap-server-definition-name'] = None

    # Add artificial property 'primary-mfa-server-definition-name'
    try:
        pmsd_uri = user.get_property('primary-mfa-server-definition-uri')
    except KeyError:
        # Property was introduced in HMC 2.15.0
        pass
    else:
        if pmsd_uri:
            # User's mfa-types includes mfa-server
            pmsd_props = client.session.get(pmsd_uri)
            properties['primary-mfa-server-definition-name'] = \
                pmsd_props['name']
        else:
            # User's mfa-types does not include mfa-server
            properties['primary-mfa-server-definition-name'] = None

    # Add artificial property 'backup-mfa-server-definition-name'
    try:
        bmsd_uri = user.get_property('backup-mfa-server-definition-uri')
    except KeyError:
        # Property was introduced in HMC 2.15.0
        pass
    else:
        if bmsd_uri:
            # User's mfa-types includes mfa-server
            bmsd_props = client.session.get(bmsd_uri)
            properties['backup-mfa-server-definition-name'] = \
                bmsd_props['name']
        else:
            # User's mfa-types does not include mfa-server
            properties['backup-mfa-server-definition-name'] = None

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_user_password(cmd_ctx, user_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)

    cmd_ctx.spinner.stop()
    password = click.prompt(
        "Enter new password for user '{u}' (at HMC {h})".
        format(u=user_name, h=cmd_ctx.session.host),
        hide_input=True, confirmation_prompt=True, type=str, err=True)
    cmd_ctx.spinner.start()

    properties = {'password': password}
    try:
        user.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Password of user '{u}' has been updated.".format(u=user_name))


def cmd_user_create(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    org_options = original_options(options)

    name_map = {
        'like': None,
        'password-rule': None,
        'ldap-server-definition': None,
        'primary-mfa-server-definition': None,
        'backup-mfa-server-definition': None,
        'mfa-type': None,
    }

    # Set up user properties from like user, if specified

    like_props = {}
    if org_options['like']:
        try:
            like_user = console.users.find_by_name(org_options['like'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)

        def _update(props, obj, name):
            """
            Update props if the obj has a property with the name.

            The purpose of this is to copy only User properties that actually
            exist on the version of the targeted HMC. For example, the
            MFA-related properties were added in HMC version 2.15.0.
            """
            try:
                value = obj.get_property(name)
            except KeyError:
                return
            props[name] = value

        # The following properties are taken from the like user. These are
        # the properties that do not contain anything about the identity of the
        # user. The following identifying properties are not used:
        # - name
        # - password
        # - description
        # - email-address
        # - mfa-userid
        # - mfa-userid-override
        like_props = {}
        _update(like_props, like_user, 'type')
        _update(like_props, like_user, 'disabled')
        _update(like_props, like_user, 'authentication-type')
        _update(like_props, like_user, 'password-rule-uri')
        _update(like_props, like_user, 'force-password-change')
        _update(like_props, like_user, 'ldap-server-definition-uri')
        _update(like_props, like_user, 'userid-on-ldap-server')
        _update(like_props, like_user, 'session-timeout')
        _update(like_props, like_user, 'verify-timeout')
        _update(like_props, like_user, 'idle-timeout')
        _update(like_props, like_user, 'min-pw-change-time')
        _update(like_props, like_user, 'max-failed-logins')
        _update(like_props, like_user, 'disable-delay')
        _update(like_props, like_user, 'inactivity-timeout')
        _update(like_props, like_user, 'disruptive-pw-required')
        _update(like_props, like_user, 'disruptive-text-required')
        _update(like_props, like_user, 'allow-remote-access')
        _update(like_props, like_user, 'allow-management-interfaces')
        _update(like_props, like_user, 'max-web-services-api-sessions')
        _update(like_props, like_user, 'web-services-api-session-idle-timeout')
        _update(like_props, like_user, 'multi-factor-authentication-required')
        # The following are only present if mfa required:
        _update(like_props, like_user, 'mfa-types')
        _update(like_props, like_user, 'primary-mfa-server-definition-uri')
        _update(like_props, like_user, 'backup-mfa-server-definition-uri')
        _update(like_props, like_user, 'mfa-policy')

    # Determine user properties from options

    option_props = options_to_properties(org_options, name_map)

    if org_options['password-rule'] in (None, ''):
        pass  # omit -> HMC sets to default
    else:
        try:
            rule = console.password_rules.find_by_name(
                org_options['password-rule'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        option_props['password-rule-uri'] = rule.uri

    if org_options['ldap-server-definition'] in (None, ''):
        pass  # omit -> HMC sets to default
    else:
        try:
            ldap_def = console.ldap_server_definitions.find_by_name(
                org_options['ldap-server-definition'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        option_props['ldap-server-definition-uri'] = ldap_def.uri

    if org_options['primary-mfa-server-definition'] in (None, ''):
        pass  # omit -> HMC sets to default
    else:
        raise NotImplementedError
        # TODO: Implement zhmcclient.MfaServerDefinition
        # try:
        #     mfa_def = console.mfa_server_definitions.find_by_name(
        #         org_options['primary-mfa-server-definition'])
        # except zhmcclient.Error as exc:
        #     raise click_exception(exc, cmd_ctx.error_format)
        # option_props['primary-mfa-server-definition-uri'] = mfa_def.uri

    if org_options['backup-mfa-server-definition'] in (None, ''):
        pass  # omit -> HMC sets to default
    else:
        raise NotImplementedError
        # TODO: Implement zhmcclient.MfaServerDefinition
        # try:
        #     mfa_def = console.mfa_server_definitions.find_by_name(
        #         org_options['backup-mfa-server-definition'])
        # except zhmcclient.Error as exc:
        #     raise click_exception(exc, cmd_ctx.error_format)
        # option_props['backup-mfa-server-definition-uri'] = mfa_def.uri

    if client.version_info() >= API_VERSION_HMC_2_15_0:
        if org_options['mfa-type'] in (None, ''):
            # 'mfa-types' remains unset in this case
            option_props['multi-factor-authentication-required'] = False
        elif org_options['mfa-type'] == 'hmc-totp':
            option_props['mfa-types'] = ['hmc-totp']
            option_props['multi-factor-authentication-required'] = True
        else:
            assert org_options['mfa-type'] == 'mfa-server'
            option_props['mfa-types'] = ['mfa-server']

        if org_options['mfa-policy'] == '':
            option_props['mfa-policy'] = None

        if org_options['mfa-userid'] == '':
            option_props['mfa-userid'] = None

        if org_options['mfa-userid-override'] == '':
            option_props['mfa-userid-override'] = None
    else:
        if org_options['mfa-type'] is not None \
                or org_options['mfa-policy'] is not None \
                or org_options['mfa-userid'] is not None \
                or org_options['mfa-userid-override'] is not None:
            raise click_exception(
                "Use of the MFA-related options --mfa-type, --mfa-policy, "
                "--mfa-userid, --mfa-userid-override require "
                "HMC version 2.15.0 or later.",
                cmd_ctx.error_format)

    if client.version_info() >= API_VERSION_HMC_2_14_0:
        if org_options['email-address'] == '':
            option_props['email-address'] = None
    else:
        if org_options['email-address'] is not None:
            raise click_exception(
                "Use of the --email-address option requires "
                "HMC version 2.14.0 or later.",
                cmd_ctx.error_format)

    properties = like_props
    properties.update(option_props)

    try:
        new_user = console.users.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New user '{u}' has been created.".
               format(u=new_user.properties['name']))


def cmd_user_update(cmd_ctx, user_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)

    org_options = original_options(options)

    name_map = {
        'password-rule': None,
        'default-group': None,
        'ldap-server-definition': None,
        'primary-mfa-server-definition': None,
        'backup-mfa-server-definition': None,
        'mfa-type': None,
    }

    properties = options_to_properties(org_options, name_map)

    if org_options['password-rule'] is None:
        pass  # omit -> no change
    elif org_options['password-rule'] == '':
        properties['password-rule-uri'] = None
    else:
        try:
            rule = console.password_rules.find_by_name(
                org_options['password-rule'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        properties['password-rule-uri'] = rule.uri

    if org_options['default-group'] is None:
        pass  # omit -> no change
    elif org_options['default-group'] == '':
        properties['default-group-uri'] = None
    else:
        raise NotImplementedError
        # TODO: Implement zhmcclient.Group
        # try:
        #     group = client.groups.find_by_name(
        #         org_options['default-group'])
        # except zhmcclient.Error as exc:
        #     raise click_exception(exc, cmd_ctx.error_format)
        # properties['default-group-uri'] = group.uri

    if org_options['ldap-server-definition'] is None:
        pass  # omit -> no change
    elif org_options['ldap-server-definition'] == '':
        properties['ldap-server-definition-uri'] = None
    else:
        try:
            ldap_def = console.ldap_server_definitions.find_by_name(
                org_options['ldap-server-definition'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        properties['ldap-server-definition-uri'] = ldap_def.uri

    if org_options['primary-mfa-server-definition'] is None:
        pass  # omit -> no change
    elif org_options['primary-mfa-server-definition'] == '':
        properties['primary-mfa-server-definition-uri'] = None
    else:
        raise NotImplementedError
        # TODO: Implement zhmcclient.MfaServerDefinition
        # try:
        #     mfa_def = console.mfa_server_definitions.find_by_name(
        #         org_options['primary-mfa-server-definition'])
        # except zhmcclient.Error as exc:
        #     raise click_exception(exc, cmd_ctx.error_format)
        # properties['primary-mfa-server-definition-uri'] = mfa_def.uri

    if org_options['backup-mfa-server-definition'] is None:
        pass  # omit -> no change
    elif org_options['backup-mfa-server-definition'] == '':
        properties['backup-mfa-server-definition-uri'] = None
    else:
        raise NotImplementedError
        # TODO: Implement zhmcclient.MfaServerDefinition
        # try:
        #     mfa_def = console.mfa_server_definitions.find_by_name(
        #         org_options['backup-mfa-server-definition'])
        # except zhmcclient.Error as exc:
        #     raise click_exception(exc, cmd_ctx.error_format)
        # properties['backup-mfa-server-definition-uri'] = mfa_def.uri

    if client.version_info() >= API_VERSION_HMC_2_15_0:
        if org_options['mfa-type'] is None:
            pass  # omit -> no change
        elif org_options['mfa-type'] == '':
            # 'mfa-types' remains unset in this case
            properties['multi-factor-authentication-required'] = False
        elif org_options['mfa-type'] == 'hmc-totp':
            properties['mfa-types'] = ['hmc-totp']
            properties['multi-factor-authentication-required'] = True
        else:
            assert org_options['mfa-type'] == 'mfa-server'
            properties['mfa-types'] = ['mfa-server']

        if org_options['mfa-policy'] == '':
            properties['mfa-policy'] = None

        if org_options['mfa-userid'] == '':
            properties['mfa-userid'] = None

        if org_options['mfa-userid-override'] == '':
            properties['mfa-userid-override'] = None
    else:
        if org_options['mfa-type'] is not None \
                or org_options['mfa-policy'] is not None \
                or org_options['mfa-userid'] is not None \
                or org_options['mfa-userid-override'] is not None:
            raise click_exception(
                "Use of the MFA-related options --mfa-type, --mfa-policy, "
                "--mfa-userid, --mfa-userid-override require "
                "HMC version 2.15.0 or later.",
                cmd_ctx.error_format)

    if client.version_info() >= API_VERSION_HMC_2_14_0:
        if org_options['email-address'] == '':
            properties['email-address'] = None
    else:
        if org_options['email-address'] is not None:
            raise click_exception(
                "Use of the --email-address option requires "
                "HMC version 2.14.0 or later.",
                cmd_ctx.error_format)

    if client.version_info() < API_VERSION_HMC_2_14_0 \
            and org_options['force-shared-secret-key-change'] is not None:
        raise click_exception(
            "Use of the --force-shared-secret-key-change option requires "
            "HMC version 2.14.0 or later.",
            cmd_ctx.error_format)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating user '{u}'.".
                   format(u=user_name))
        return

    try:
        user.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User '{u}' has been updated.".format(u=user_name))


def cmd_user_delete(cmd_ctx, user_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)

    try:
        user.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User '{u}' has been deleted.".format(u=user_name))


def cmd_user_add_role(cmd_ctx, user_name, user_role_name):
    # pylint: disable=missing-function-docstring

    # pylint: disable=import-outside-toplevel,cyclic-import
    from ._cmd_user_role import find_user_role

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    try:
        user.add_user_role(user_role)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User role '{r}' has been added to user '{u}'.".
               format(r=user_role_name, u=user_name))


def cmd_user_remove_role(cmd_ctx, user_name, user_role_name):
    # pylint: disable=missing-function-docstring

    # pylint: disable=import-outside-toplevel,cyclic-import
    from ._cmd_user_role import find_user_role

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user = find_user(cmd_ctx, console, user_name)
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    try:
        user.remove_user_role(user_role)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User role '{r}' has been removed from user '{u}'.".
               format(r=user_role_name, u=user_name))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for adapters.
"""

from __future__ import absolute_import

import time
import re
from datetime import datetime
import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS
from ._cmd_cpc import find_cpc


# Mapping of --crypto-type option values to crypto-type adapter property values
CRYPTO_TYPE_PROPERTIES = {
    'acc': 'accelerator',
    'cca': 'cca-coprocessor',
    'ep11': 'ep11-coprocessor',
}

# List of --crypto-type option values
CRYPTO_TYPE_OPTIONS = sorted(list(CRYPTO_TYPE_PROPERTIES.keys()))

# Default crypto adapter name patterns
CRYPTO_DEFAULT_NAMES = re.compile(
    r"^(Accel|CCA|EP11) [0-9A-F]{4} [0-9A-F]{4}-[0-9]{2}$")

# Mapping of --storage-type option values to type adapter property values
STORAGE_TYPE_PROPERTIES = {
    'fcp': 'fcp',
    'fc': 'fc',
    'not-configured': 'not-configured',
}

# List of --storage-type option values
STORAGE_TYPE_OPTIONS = sorted(list(STORAGE_TYPE_PROPERTIES.keys()))


def find_adapter(cmd_ctx, client, cpc_or_name, adapter_name):
    """
    Find an adapter by name and return its resource object.
    """
    if isinstance(cpc_or_name, zhmcclient.Cpc):
        cpc = cpc_or_name
    else:
        cpc = find_cpc(cmd_ctx, client, cpc_or_name)
    # The CPC must be in DPM mode. We don't check that because it would
    # cause a GET to the CPC resource that we otherwise don't need.
    try:
        adapter = cpc.adapters.find(name=adapter_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return adapter


@cli.group('adapter', options_metavar=COMMAND_OPTIONS_METAVAR)
def adapter_group():
    """
    Command group for managing adapters (DPM mode only).

    Physical adapters (e.g. OSA, FICON) are auto-discovered and cannot be
    created. Logical adapters (HiperSockets) can be created and deleted.

    The commands in this group work only on CPCs in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@adapter_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--type', is_flag=True, required=False, hidden=True)
@add_options(LIST_OPTIONS)
@click.pass_obj
def adapter_list(cmd_ctx, cpc, **options):
    """
    List the adapters in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_adapter_list(cmd_ctx, cpc, options))


@adapter_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@click.pass_obj
def adapter_show(cmd_ctx, cpc, adapter):
    """
    Show the details of an adapter.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent CPC.
      - 'network-port-names' - Names of the Adapter Ports referenced by
        'network-port-uris' if present (for network adapters, index-correlated).
      - 'network-port-indexes' - Indexes of the Adapter Ports referenced by
        'network-port-uris' if present (for network adapters, index-correlated).
      - 'storage-port-names' - Names of the Adapter Ports referenced by
        'storage-port-uris' if present (for storage adapters, index-correlated).
      - 'storage-port-indexes' - Indexes of the Adapter Ports referenced by
        'storage-port-uris' if present (for storage adapters, index-correlated).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_adapter_show(cmd_ctx, cpc, adapter))


@adapter_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@click.option('--name', type=str, required=False,
              help='The new name of the adapter.')
@click.option('--description', type=str, required=False,
              help='The new description of the adapter.')
@click.option('--port-description', type=str, required=False,
              help='The new description of the single port of the adapter.')
@click.option('--mtu-size', type=click.Choice(['8', '16', '32', '56']),
              required=False,
              help='The new MTU size of the adapter in KiB. '
              '(HiperSockets only).')
@click.option('--allowed-capacity', type=int, required=False,
              help='The new maximum number of HBAs per partition. '
              '(only certain adapter types).')
@click.option('--chpid', type=str, required=False,
              help='The new channel path ID (CHPID, 2 hex chars) used by the '
              'adapter\'s partition. (OSA, FICON, HiperSockets only; Adapter '
              'must not be used by any partition).')
@click.option('--crypto-tke/--no-crypto-tke', default=None, required=False,
              help='The new setting whether TKE commands are permitted on the '
              'crypto adapter. (Crypto only).')
@click.option('--crypto-type', default=None, required=False,
              type=click.Choice(CRYPTO_TYPE_OPTIONS),
              help='The new crypto type (Crypto only; Adapter must be '
              'offline).')
@click.option('--storage-type', default=None, required=False,
              type=click.Choice(STORAGE_TYPE_OPTIONS),
              help='The new ficon type (FICON only; Adapter must be '
              'offline).')
@click.option('--dont-zeroize', is_flag=True, default=False, required=False,
              help='Do not zeroize when changing crypto type to '
              'accelerator. (Crypto only; Default is to zeroize).')
@click.pass_obj
def adapter_update(cmd_ctx, cpc, adapter, **options):
    """
    Update the properties of an adapter.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    The adapter may be a physical adapter (e.g. a discovered OSA card) or a
    logical adapter (e.g. HiperSockets).

    For Crypto Express adapters, the crypto type can be changed.

    For FICON Express adapters, the ficon type can be changed.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_adapter_update(cmd_ctx, cpc, adapter,
                                                   options))


@adapter_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--name', type=str, required=True,
              help='The name of the new adapter.')
@click.option('--description', type=str, required=False,
              help='The description of the new adapter. Default: empty')
@click.option('--port-description', type=str, required=False,
              help='The description of the (single) port of the new adapter. '
              'Default: empty')
@click.option('--mtu-size', type=click.Choice(['8', '16', '32', '56']),
              required=False,
              help='The MTU size of the new adapter in KiB. Default: 8')
@click.pass_obj
def adapter_create_hipersocket(cmd_ctx, cpc, **options):
    """
    Create a HiperSockets adapter in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.

    Some more properties of the new HiperSockets adapter can be set via
    adapter update.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_adapter_create_hipersocket(cmd_ctx, cpc, options))


@adapter_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the adapter.',
              prompt='Are you sure you want to delete this adapter ?')
@click.pass_obj
def adapter_delete_hipersocket(cmd_ctx, cpc, adapter):
    """
    Delete a HiperSockets adapter.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_adapter_delete_hipersocket(cmd_ctx, cpc, adapter))


@adapter_group.command('list-nics', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@add_options(LIST_OPTIONS)
@click.pass_obj
def adapter_list_nics(cmd_ctx, cpc, adapter, **options):
    """
    List the NICs backed by a network adapter in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_adapter_list_nics(cmd_ctx, cpc, adapter, options))


def cmd_adapter_list(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    try:
        adapters = cpc.adapters.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['type']:
        click.echo("The --type option is deprecated and type information "
                   "is now always shown.")

    show_list = [
        'name',
        'cpc',
    ]
    if not options['names_only']:
        show_list.extend([
            'adapter-id',
            'status',
            'adapter-family',
            'type',
            'detected-card-type',
            'crypto-type',
            'card-location',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    cpc_additions = {}
    for adapter in adapters:
        cpc_additions[adapter.uri] = cpc_name
    additions = {
        'cpc': cpc_additions,
    }

    try:
        print_resources(cmd_ctx, adapters, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_adapter_show(cmd_ctx, cpc_name, adapter_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)

    try:
        adapter.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(adapter.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = cpc_name

    # Add artificial properties 'network-port-names', 'network-port-indexes'
    try:
        netport_uris = adapter.properties['network-port-uris']
    except KeyError:
        pass
    else:
        # It is a network adapter
        netport_names = []
        netport_indexes = []
        for netport_uri in netport_uris:
            netport_props = client.session.get(netport_uri)
            netport_names.append(netport_props['name'])
            netport_indexes.append(netport_props['index'])
        properties['network-port-names'] = netport_names
        properties['network-port-indexes'] = netport_indexes

    # Add artificial properties 'storage-port-names', 'storage-port-indexes'
    try:
        stoport_uris = adapter.properties['storage-port-uris']
    except KeyError:
        pass
    else:
        # It is a storage adapter
        stoport_names = []
        stoport_indexes = []
        for stoport_uri in stoport_uris:
            # For unconfigurd FICON adapters, the 'storage-port-uris' property
            # is set, but Get Adapter Properties fails with 404,4
            try:
                stoport_props = client.session.get(stoport_uri)
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 404 and exc.reason == 4:
                    # Unconfigured FICON adapter
                    pass
                else:
                    raise
            else:
                stoport_names.append(stoport_props['name'])
                stoport_indexes.append(stoport_props['index'])
        properties['storage-port-names'] = stoport_names
        properties['storage-port-indexes'] = stoport_indexes

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_adapter_update(cmd_ctx, cpc_name, adapter_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)
    new_name = None  # New name if adapter name has changed

    name_map = {
        'mtu-size': 'maximum-transmission-unit-size',
        'chpid': 'channel-path-id',
        'crypto-tke': 'tke-commands-enabled',
        'crypto-type': None,
        'dont-zeroize': None,
        'storage-type': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)
    crypto_type_option = options['crypto_type']
    storage_type_option = options['storage_type']

    if not properties and not crypto_type_option and not storage_type_option:
        cmd_ctx.spinner.stop()
        click.echo("No changes specified for updating adapter '{a}'.".
                   format(a=adapter_name))
        return

    if 'name' in properties and properties['name'] != adapter_name:
        new_name = properties['name']
    try:
        adapter.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if crypto_type_option:
        crypto_type_prop = CRYPTO_TYPE_PROPERTIES[crypto_type_option]
        if crypto_type_option == 'acc':
            zeroize = not options['dont_zeroize']
        else:
            if options['dont_zeroize']:
                cmd_ctx.spinner.stop()
                click.echo("Ignoring --dont-zeroize option for adapter '{a}'.".
                           format(a=adapter_name))
            zeroize = None
        try:
            adapter.change_crypto_type(crypto_type_prop, zeroize)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)

        # Changing the crypto type triggers a change of the adapter name,
        # if the current name is the default name (e.g. 'CCA 0104 A14B-03')
        # The new name will be visible at the API only after about 2 seconds.
        if CRYPTO_DEFAULT_NAMES.match(adapter.name):

            # This debug code can be used to test whether the behavior of
            # changing the adapter name has changed.
            DEBUG_CRYPTO_NAME_CHANGE = False

            if DEBUG_CRYPTO_NAME_CHANGE:
                t1 = datetime.now()
                cmd_ctx.spinner.stop()
                for _ in range(0, 20):
                    adapter.pull_full_properties()
                    dt = datetime.now() - t1
                    print("Debug: Adapter name after {}: {!r}".
                          format(dt, adapter.name))
            else:
                time.sleep(4)
                adapter.pull_full_properties()

            if adapter.name != adapter_name:
                new_name = adapter.name

    if storage_type_option:
        storage_type_prop = STORAGE_TYPE_PROPERTIES[storage_type_option]
        try:
            adapter.change_adapter_type(storage_type_prop)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        sibling_adapters = adapter.list_sibling_adapters()
        sibling_adapter_names = ', '.join(
            ["'{}'".format(a.name) for a in sibling_adapters])
        cmd_ctx.spinner.stop()
        click.echo("FICON adapter {a} and its sibling adapter {s} have "
                   "been changed to type '{t}'.".
                   format(a=adapter.name, s=sibling_adapter_names,
                          t=storage_type_option))

    cmd_ctx.spinner.stop()
    if new_name:
        click.echo("Adapter '{a}' has been renamed to '{n}' and was updated.".
                   format(a=adapter_name, n=new_name))
    else:
        click.echo("Adapter '{a}' has been updated.".format(a=adapter_name))


def cmd_adapter_create_hipersocket(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name_map = {
        'mtu-size': 'maximum-transmission-unit-size',
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    try:
        new_adapter = cpc.adapters.create_hipersocket(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New HiperSockets adapter '{a}' has been created.".
               format(a=new_adapter.properties['name']))


def cmd_adapter_delete_hipersocket(cmd_ctx, cpc_name, adapter_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)

    try:
        adapter.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("HiperSockets adapter '{a}' has been deleted.".
               format(a=adapter_name))


def cmd_adapter_list_nics(cmd_ctx, cpc_name, adapter_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)

    adapter_family = adapter.get_property('adapter-family')
    adapter_type = adapter.get_property('type')

    if adapter_family not in ('hipersockets', 'osa', 'roce', 'cna'):
        raise click_exception(
            "Adapter {f!r} on CPC {c} is not a network adapter".
            format(f=adapter.name, c=cpc_name), cmd_ctx.error_format)

    if adapter_type == 'osm':
        raise click_exception(
            "Adapter {f!r} on CPC {c} is an OSM adapter and cannot be "
            "assigned to partitions".
            format(f=adapter.name, c=cpc_name), cmd_ctx.error_format)

    results = []  # tuple(partition, nic, port_index)
    partitions = adapter.list_assigned_partitions()
    for part in partitions:
        nics = part.nics.list()
        for nic in nics:
            try:
                vswitch_uri = nic.get_property('virtual-switch-uri')
            except KeyError:
                pass
            else:
                vswitch_props = client.session.get(vswitch_uri)
                backing_adapter_uri = vswitch_props['backing-adapter-uri']
                backing_port_index = vswitch_props['port']
            try:
                port_uri = nic.get_property('network-adapter-port-uri')
            except KeyError:
                pass
            else:
                port_props = client.session.get(port_uri)
                backing_adapter_uri = port_props['parent']
                backing_port_index = port_props['index']

            if backing_adapter_uri == adapter.uri:
                results.append((part, nic, backing_port_index))

    show_list = [
        'name',
        'partition',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'port-index',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    partition_additions = {}
    port_index_additions = {}
    for part, nic, port_index in results:
        partition_additions[nic.uri] = part.name
        port_index_additions[nic.uri] = port_index
    additions = {
        'partition': partition_additions,
        'port-index': port_index_additions,
    }

    nics = [nic for _, nic, _ in results]
    try:
        print_resources(cmd_ctx, nics, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC sessions.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import click_exception


@cli.group('session')
def session_group():
    """
    Command group for managing sessions.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified before the
    command.
    """


@session_group.command('create')
@click.pass_obj
def session_create(cmd_ctx):
    """
    Create an HMC session.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified before the
    command.
    """
    cmd_ctx.execute_cmd(lambda: cmd_session_create(cmd_ctx), logoff=False)


@session_group.command('delete')
@click.pass_obj
def session_delete(cmd_ctx):
    """
    Delete the current HMC session.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified before the
    command.
    """
    cmd_ctx.execute_cmd(lambda: cmd_session_delete(cmd_ctx))


def cmd_session_create(cmd_ctx):
    """Create an HMC session."""
    session = cmd_ctx.session
    try:
        # We need to first log off, to make the logon really create a new
        # session. If we don't first log off, the session from the
        # ZHMC_SESSION_ID env var will be used and no new session be created.
        session.logoff()
        session.logon(verify=True)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()

    if session.verify_cert is False:
        no_verify = 'TRUE'
        ca_certs = None
    elif session.verify_cert is True:
        no_verify = None
        ca_certs = None
    else:
        no_verify = None
        ca_certs = session.verify_cert

    click.echo("export ZHMC_HOST={h}".format(h=session.host))
    click.echo("export ZHMC_USERID={u}".format(u=session.userid))
    click.echo("export ZHMC_SESSION_ID={s}".format(s=session.session_id))
    if no_verify is None:
        click.echo("unset ZHMC_NO_VERIFY")
    else:
        click.echo("export ZHMC_NO_VERIFY={nv}".format(nv=no_verify))
    if ca_certs is None:
        click.echo("unset ZHMC_CA_CERTS")
    else:
        click.echo("export ZHMC_CA_CERTS={cc}".format(cc=ca_certs))


def cmd_session_delete(cmd_ctx):
    """Delete the current HMC session."""
    session = cmd_ctx.session
    try:
        session.logoff()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("unset ZHMC_SESSION_ID")
